#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#define BACK_LOG            10

void* getClientSockAddr(struct sockaddr*);

int main(int argc, char** argv) {
    /* Usage: server <port> */
    if(argc != 2) {
        fprintf(stderr, "Usage: %s <port>\n", argv[0]);
        return 1;
    }

    //int msgLen = strlen(argv[2]);

    struct addrinfo hints = {};
    struct addrinfo *servAddrInfo;  
    hints.ai_family = AF_UNSPEC;
    hints.ai_flags = AI_PASSIVE;
    hints.ai_socktype = SOCK_STREAM;

    int status = getaddrinfo(NULL, argv[1], &hints, &servAddrInfo);
    if(0 != status) {
        fprintf(stderr, "getaddrinfo(): %s\n", strerror(status));
        return status;
    }

    struct addrinfo *ptr;
    int listenSockfd;
    for(ptr = servAddrInfo;
        ptr != NULL; 
        ptr = ptr->ai_next) {
            listenSockfd = socket(ptr->ai_family, 
                                ptr->ai_socktype, 
                                ptr->ai_protocol);
            int tmp = errno;
            if(-1 == listenSockfd) {
                fprintf(stderr, "socket(): %s\n", strerror(tmp));
                continue;
            }

            int optvalue = 1;
            int status = setsockopt(listenSockfd, 
                                    SOL_SOCKET, 
                                    SO_REUSEADDR, 
                                    &optvalue, 
                                    sizeof(optvalue));
            if(status == -1) {
                int tmp = errno;
                fprintf(stderr, "setsockopt(): %s\n", strerror(tmp));
                close(listenSockfd);
                return tmp;
            }

            status = setsockopt(listenSockfd, 
                                    SOL_SOCKET, 
                                    SO_REUSEPORT, 
                                    &optvalue, 
                                    sizeof(optvalue));
            if(status == -1) {
                int tmp = errno;
                fprintf(stderr, "setsockopt(): %s\n", strerror(tmp));
                close(listenSockfd);
                return tmp;
            }

            status = bind(listenSockfd, ptr->ai_addr, ptr->ai_addrlen);
            if(-1 == status) {
                close(listenSockfd);
                fprintf(stderr, "bind(): %s\n", strerror(status));
                continue;
            }

            break;
    }

    freeaddrinfo(servAddrInfo);
    if(ptr == NULL) {
        fprintf(stderr, "Local address unavailable.\n");
        return 1;
    }

    if(-1 == listen(listenSockfd, BACK_LOG)) {
        int tmp = errno;
        fprintf(stderr, "listen(): %s\n", strerror(tmp));
        return tmp;
    }

    int connSockfd;
    struct sockaddr_storage clientSockAddr = {};
    socklen_t clientSockAddrLen;
    while(1) {
        clientSockAddrLen = sizeof(struct sockaddr_storage);
        connSockfd = accept(listenSockfd, 
                            &clientSockAddr, 
                            &clientSockAddrLen);
        if(connSockfd == -1) 
            continue;
        
        char* clientAddrStr[INET6_ADDRSTRLEN];
        inet_ntop(clientSockAddr.ss_family,
                  getClientSockAddr((struct sockaddr*)&clientSockAddr),
                  clientAddrStr,
                  INET6_ADDRSTRLEN);
        //fprintf(stdout, "accept(): connection from %s\n", clientAddrStr);
        if(send(connSockfd, clientAddrStr, strlen(clientAddrStr), 0) == -1) {
            int tmp = errno;
            fprintf(stderr, "send(): %s\n", strerror(tmp));
            close(connSockfd);
            continue;
        }
        close(connSockfd);
    }

    return 0;
}

void* getClientSockAddr(struct sockaddr* client) {
    if(client->sa_family == AF_INET)
        return &(((struct sockaddr_in*)client)->sin_addr);
    else
        return &(((struct sockaddr_in6*)client)->sin6_addr);
}
